import os
import re
import time
import unicodedata
from collections import defaultdict, deque
from contextlib import contextmanager
from itertools import chain, islice, tee
from pathlib import Path
from statistics import fmean
from typing import Deque, Dict, Iterable, Optional, Tuple, TypeVar, Union

import click

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


def _format_size(value: float, digits: int, unit: str):
    if digits > 0:
        return f"{{:.{digits}f}}{unit}".format(value)
    else:
        return f"{int(value)}{unit}"


def format_size(bytes_: Union[int, float], digits: int = 1):
    if bytes_ < 1024:
        return _format_size(bytes_, digits, "B")

    kilo = bytes_ / 1024
    if kilo < 1024:
        return _format_size(kilo, digits, "kB")

    mega = kilo / 1024
    if mega < 1024:
        return _format_size(mega, digits, "MB")

    return _format_size(mega / 1024, digits, "GB")


def format_duration(total_seconds: Union[int, float]) -> str:
    total_seconds = int(total_seconds)
    hours = total_seconds // 3600
    remainder = total_seconds % 3600
    minutes = remainder // 60
    seconds = total_seconds % 60

    if hours:
        return f"{hours} h {minutes} min"

    if minutes:
        return f"{minutes} min {seconds} sec"

    return f"{seconds} sec"


def format_time(total_seconds: Union[int, float], force_hours: bool = False) -> str:
    total_seconds = int(total_seconds)
    hours = total_seconds // 3600
    remainder = total_seconds % 3600
    minutes = remainder // 60
    seconds = total_seconds % 60

    if hours or force_hours:
        return f"{hours:02}:{minutes:02}:{seconds:02}"

    return f"{minutes:02}:{seconds:02}"


def read_int(msg: str, min: int, max: int, default: Optional[int] = None) -> int:
    while True:
        try:
            val = click.prompt(msg, default=default, type=int)
            if default and not val:
                return default
            if min <= int(val) <= max:
                return int(val)
        except ValueError:
            pass


def slugify(value: str) -> str:
    value = unicodedata.normalize("NFKC", str(value))
    value = re.sub(r"[^\w\s_-]", "", value)
    value = re.sub(r"[\s_-]+", "_", value)
    return value.strip("_").lower()


def titlify(value: str) -> str:
    value = unicodedata.normalize("NFKC", str(value))
    value = re.sub(r"[^\w\s\[\]().-]", "", value)
    value = re.sub(r"\s+", " ", value)
    return value.strip()


VIDEO_ID_PATTERN = r"(?P<id>\d+)"

VIDEO_PATTERNS = [
    rf"^{VIDEO_ID_PATTERN}$",
    rf"^https://(www\.|m\.)?twitch\.tv/videos/{VIDEO_ID_PATTERN}(\?.+)?$",
    rf"^https://(www\.|m\.)?twitch\.tv/\w+/video/{VIDEO_ID_PATTERN}(\?.+)?$",
]

CLIP_SLUG_PATTERN = r"(?P<slug>[A-Za-z0-9-_]+)"

CLIP_PATTERNS = [
    rf"^{CLIP_SLUG_PATTERN}$",
    rf"^https://(www\.|m\.)?twitch\.tv/\w+/clip/{CLIP_SLUG_PATTERN}(\?.+)?$",
    rf"^https://clips\.twitch\.tv/{CLIP_SLUG_PATTERN}(\?.+)?$",
]


def parse_video_identifier(identifier: str) -> Optional[str]:
    """Given a video ID or URL returns the video ID, or null if not matched"""
    for pattern in VIDEO_PATTERNS:
        match = re.match(pattern, identifier)
        if match:
            return match.group("id")


def parse_clip_identifier(identifier: str) -> Optional[str]:
    """Given a clip slug or URL returns the clip slug, or null if not matched"""
    for pattern in CLIP_PATTERNS:
        match = re.match(pattern, identifier)
        if match:
            return match.group("slug")


def remove_null_values(adict: Dict[K, V]) -> Dict[K, V]:
    return {k: v for k, v in adict.items() if v is not None}


def iterate_with_next(iterable: Iterable[T]) -> Iterable[Tuple[T, Optional[T]]]:
    """
    Creates an iterator which provides the current and next item.
    """
    items, nexts = tee(iterable, 2)
    nexts = chain(islice(nexts, 1, None), [None])
    return zip(items, nexts)


@contextmanager
def timed(message: str = "Time taken"):
    """Print how long it took to execute a block of code"""
    start = time.monotonic()
    yield
    duration = time.monotonic() - start
    print(f"{message}: {int(1000 * duration):.1f}ms")


perfs: Dict[str, Deque[float]] = defaultdict(lambda: deque(maxlen=50))


@contextmanager
def monitor_performance(group: str):
    from twitchdl.output import print_status

    global perfs
    start = time.monotonic()
    yield
    perfs[group].append(time.monotonic() - start)
    print_status(f"{group}: {1000 * fmean(perfs[group]):.1f}ms")


def get_size(path: Path):
    """Get the total size of all files in the given directory."""
    if path.is_file():
        return os.path.getsize(path)

    size = 0
    for subpath, _, filenames in os.walk(path):
        for filename in filenames:
            full_path = os.path.join(subpath, filename)
            size += os.path.getsize(full_path)
    return size



from twitchdl.playlists import Vod, filter_vods


def test_filter_vods():
    vods = [Vod(index, path=f"vod_{index}.ts", duration=10.0) for index in range(100)]

    # No cropping required
    filtered_vods, crop_start, crop_duration = filter_vods(vods, 60, 120)
    assert filtered_vods[0].index == 6
    assert filtered_vods[-1].index == 11
    assert crop_start is None
    assert crop_duration is None

    # Croping required at start
    filtered_vods, crop_start, crop_duration = filter_vods(vods, 63, 120)
    assert filtered_vods[0].index == 6
    assert filtered_vods[-1].index == 11
    assert crop_start == 3
    assert crop_duration is None

    # Croping required at end
    filtered_vods, crop_start, crop_duration = filter_vods(vods, 60, 115)
    assert filtered_vods[0].index == 6
    assert filtered_vods[-1].index == 11
    assert crop_start is None
    assert crop_duration == 55

    # Croping required at start and end
    filtered_vods, crop_start, crop_duration = filter_vods(vods, 63, 115)
    assert filtered_vods[0].index == 6
    assert filtered_vods[-1].index == 11
    assert crop_start == 3
    assert crop_duration == 52



import hashlib
import logging
import os
import sys
from pathlib import Path
from typing import List, Optional

from httpx import HTTPError

from twitchdl.http import download_file
from twitchdl.output import print_error, print_status

CACHE_SUBFOLDER = "twitch-dl"


logger = logging.getLogger(__name__)


def download_cached(
    url: str,
    *,
    subdir: Optional[str] = None,
    filename: Optional[str] = None,
) -> Path:
    target_dir = get_cache_dir(subdir)

    if not filename:
        filename = hashlib.sha256(url.encode()).hexdigest()
    target = target_dir / filename

    if not target.exists():
        print_status(f"Downloading {url}", dim=True)
        download_file(url, target)

    return target


def download_cached_or_none(
    url: str,
    *,
    subdir: Optional[str] = None,
    filename: Optional[str] = None,
) -> Optional[Path]:
    try:
        return download_cached(url, subdir=subdir, filename=filename)
    except HTTPError as ex:
        print_error(ex)
        return None


def get_cache_dir(subdir: Optional[str] = None) -> Path:
    path = _cache_dir_path()
    if subdir:
        path = path / subdir
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_cache_subdirs() -> List[Path]:
    subdirs: List[Path] = []
    for item in _cache_dir_path().iterdir():
        if item.is_dir():
            subdirs.append(item)
    return subdirs


def _cache_dir_path() -> Path:
    """Returns the path to the cache directory"""

    # Windows
    if sys.platform == "win32" and "LOCALAPPDATA" in os.environ:
        return Path(os.environ["LOCALAPPDATA"], CACHE_SUBFOLDER)

    # Mac OS
    if sys.platform == "darwin":
        return Path.home() / "Library" / "Caches" / CACHE_SUBFOLDER

    # Respect XDG_CONFIG_HOME env variable if set
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    if "XDG_CACHE_HOME" in os.environ:
        return Path(os.environ["XDG_CACHE_HOME"], CACHE_SUBFOLDER)

    return Path.home() / ".cache" / CACHE_SUBFOLDER


class Cache:
    """Helps keep track of cached files and folders and delete them when finished"""

    def __init__(self, root: Path):
        self.root = root
        self.files: List[Path] = []
        self.dirs: List[Path] = []
        self.mkdir(root)

    def get_path(self, filename: str) -> Path:
        path = self.root / filename
        self.files.append(path)
        return path

    def mkdir(self, path: Path):
        """Create a new directory recursively, save created dirs to self.dirs.
        Original: https://github.com/python/cpython/blob/3.13/Lib/pathlib/_local.py#L717"""
        try:
            os.mkdir(path)
            self.dirs.append(path)
        except FileNotFoundError:
            if path.parent == path:
                raise
            self.mkdir(path.parent)
            self.mkdir(path)
        except OSError:
            if not path.is_dir():
                raise

    def delete(self):
        try:
            for file in self.files:
                if file.exists():
                    os.remove(file)
            for dir in reversed(self.dirs):
                os.rmdir(dir)
        except Exception as ex:
            print_error(f"Failed deleting cache: {ex}\nSome files are left over in {self.root}")



import asyncio
import platform
import re
import shlex
import subprocess
from enum import Enum, auto
from pathlib import Path
from typing import Any, List, Optional
from urllib.parse import urlencode

import click
import httpx

from twitchdl import twitch, utils
from twitchdl.cache import Cache
from twitchdl.entities import Clip, DownloadOptions
from twitchdl.exceptions import ConsoleError
from twitchdl.http import download_all, download_file
from twitchdl.naming import clip_filename, video_filename, video_placeholders
from twitchdl.output import blue, bold, green, print_error, print_log, underlined, yellow
from twitchdl.playlists import (
    Playlist,
    enumerate_vods,
    filter_vods,
    get_init_sections,
    load_m3u8,
    make_join_playlist,
    parse_playlists,
    select_playlist,
)
from twitchdl.twitch import Chapter, ClipAccessToken, Video


def download(ids: List[str], args: DownloadOptions):
    if not ids:
        print_log("No IDs to downlad given")
        return

    for video_id in ids:
        download_one(video_id, args)


def download_one(id_or_slug: str, args: DownloadOptions):
    video_id = utils.parse_video_identifier(id_or_slug)
    if video_id:
        print_log("Looking up video...")
        video = twitch.get_video(video_id)
        if video:
            _download_video(video, args)
        else:
            print_error(f"Video '{video_id}' not found")
        return

    slug = utils.parse_clip_identifier(id_or_slug)
    if slug:
        print_log("Looking up clip...")
        clip = twitch.get_clip(slug)

        if clip:
            _download_clip(clip, args)
        else:
            print_error(f"Clip '{slug}' not found")
        return

    print_error(f"Not a valid video ID or clip slug: {id_or_slug}")


def _join_vods(
    playlist_path: Path,
    metadata_path: Path,
    target: Path,
    overwrite: bool,
    crop_start: Optional[float],
    crop_duration: Optional[float],
):
    command: List[str] = []

    def append(*vars: Any):
        for var in vars:
            command.append(str(var))

    append("ffmpeg")
    append("-i", playlist_path)

    # Cropping start is done before metadata, cropping duration after metadata.
    # See: https://github.com/ihabunek/twitch-dl/issues/166
    if crop_start:
        append("-ss", utils.format_time(crop_start))

    append("-i", metadata_path)
    append("-map_metadata", 1)

    if crop_duration:
        append("-t", utils.format_time(crop_duration))

    append("-c", "copy")
    append("-stats")
    append("-loglevel", "warning")
    append(f"file:{target}")

    if overwrite:
        append("-y")

    click.secho(f"{shlex.join(command)}", dim=True)
    result = subprocess.run(command)
    if result.returncode != 0:
        raise ConsoleError("Joining files failed")


def _concat_vods(vod_paths: List[Path], target: Path):
    tool = "type" if platform.system() == "Windows" else "cat"
    command = [tool] + [str(p) for p in vod_paths]

    with open(target, "wb") as target_file:
        result = subprocess.run(command, stdout=target_file)
        if result.returncode != 0:
            raise ConsoleError(f"Joining files failed: {result.stderr}")


def _get_clip_url(access_token: ClipAccessToken, quality: Optional[str]) -> str:
    qualities = access_token["videoQualities"]

    # Quality given as an argument
    if quality:
        if quality == "source":
            return qualities[0]["sourceURL"]

        selected_quality = quality.rstrip("p")  # allow 720p as well as 720
        for q in qualities:
            if q["quality"] == selected_quality:
                return q["sourceURL"]

        available = ", ".join([str(q["quality"]) for q in qualities])
        msg = f"Quality '{quality}' not found. Available qualities are: {available}"
        raise ConsoleError(msg)

    # Ask user to select quality
    click.echo("\nAvailable qualities:")
    for n, q in enumerate(qualities):
        click.echo(f"{n + 1}) {bold(q['quality'])} [{q['frameRate']} fps]")
    click.echo()

    no = utils.read_int("Choose quality", min=1, max=len(qualities), default=1)
    selected_quality = qualities[no - 1]
    return selected_quality["sourceURL"]


def get_clip_authenticated_url(slug: str, quality: Optional[str]):
    print_log("Fetching access token...")
    access_token = twitch.get_clip_access_token(slug)

    if not access_token:
        raise ConsoleError(f"Access token not found for slug '{slug}'")

    url = _get_clip_url(access_token, quality)

    query = urlencode(
        {
            "sig": access_token["playbackAccessToken"]["signature"],
            "token": access_token["playbackAccessToken"]["value"],
        }
    )

    return f"{url}?{query}"


def _download_clip(clip: Clip, args: DownloadOptions) -> None:
    target = Path(clip_filename(clip, args.output))
    _print_found_clip(clip)
    print_log(f"Target: {target}")

    if target.exists():
        if args.skip_existing:
            click.echo(f"Clip already downloaded: {green(target)}")
            return

        if not args.overwrite:
            response = _prompt_overwrite()
            if response == Overwrite.OVERWRITE_ALL:
                args.overwrite = True
            elif response == Overwrite.SKIP:
                click.echo(f"Skipping clip: {green(target)}")
                return
            elif response == Overwrite.SKIP_ALL:
                args.skip_existing = True
                click.echo(f"Skipping clip: {green(target)}")
                return
            elif response == Overwrite.ABORT:
                raise click.Abort()

    url = get_clip_authenticated_url(clip["slug"], args.quality)
    print_log(f"Downloading from: {url}")

    if args.dry_run:
        click.echo("Dry run, clip not downloaded.")
    else:
        download_file(url, target)
        click.echo(f"Downloaded clip: {green(target)}")


def _print_found_clip(clip: Clip):
    print_log(
        "Found clip:",
        green(clip["title"]),
        "by",
        yellow(clip["broadcaster"]["displayName"]),
        "playing",
        blue(clip["game"]["name"] if clip["game"] else "Unknown"),
        f"({utils.format_time(clip['durationSeconds'])})",
    )


def _download_video(video: Video, args: DownloadOptions) -> None:
    target = Path(video_filename(video, args.format, args.output))
    _print_found_video(video)
    print_log(f"Target: {blue(target)}")

    overwrite = args.overwrite
    if target.exists():
        if args.skip_existing:
            click.echo(f"Video already downloaded: {green(target)}")
            return

        if not overwrite:
            response = _prompt_overwrite()
            if response == Overwrite.OVERWRITE:
                overwrite = True
            elif response == Overwrite.OVERWRITE_ALL:
                overwrite = True
                args.overwrite = True
            elif response == Overwrite.SKIP:
                click.echo(f"Skipping clip: {green(target)}")
                return
            elif response == Overwrite.SKIP_ALL:
                args.skip_existing = True
                click.echo(f"Skipping clip: {green(target)}")
                return
            elif response == Overwrite.ABORT:
                raise click.Abort()

    print_log("Fetching chapters...")
    chapters = twitch.get_video_chapters(video["id"])
    start, end = _determine_time_range(chapters, args)

    print_log("Fetching access token...")
    access_token = twitch.get_access_token(video["id"], auth_token=args.auth_token)

    print_log("Fetching playlists...")
    playlists_text = twitch.get_playlists(video["id"], access_token)
    playlists = parse_playlists(playlists_text)
    playlist = select_playlist(playlists, args.quality)
    base_uri = re.sub("/[^/]+$", "/", playlist.url)

    print_log("Fetching playlist...")
    vods_text = http_get(playlist.url)
    vods_m3u8 = load_m3u8(vods_text)
    all_vods = enumerate_vods(vods_m3u8)
    vods, crop_start, crop_duration = filter_vods(all_vods, start, end)

    if args.dry_run:
        click.echo("Dry run, video not downloaded.")
        return

    cache_dir = _get_cache_dir(video, playlist, args)
    cache = Cache(cache_dir)
    print_log(f"Downloading to cache: {cache_dir}")

    # Create ffmpeg metadata file
    metadata_path = cache.get_path("metadata.txt")
    _write_metadata(video, chapters, metadata_path, start, end)

    # Save playlists for debugging purposes
    playlists_path = cache.get_path("playlists.m3u8")
    with open(playlists_path, "w") as f:
        f.write(playlists_text)

    playlist_path = cache.get_path("playlist.m3u8")
    with open(playlist_path, "w") as f:
        f.write(vods_text)

    init_sections = get_init_sections(vods_m3u8)
    for uri in init_sections:
        print_log(f"Downloading init section {uri}...")
        init_section_path = cache.get_path(uri)
        download_file(f"{base_uri}{uri}", init_section_path)

    print_log(f"Downloading {len(vods)} VODs using {args.max_workers} workers")

    sources = [base_uri + vod.path for vod in vods]
    targets = [cache.get_path(f"{vod.index:05d}.ts") for vod in vods]

    asyncio.run(
        download_all(
            zip(sources, targets),
            args.max_workers,
            rate_limit=args.rate_limit,
            count=len(vods),
        )
    )

    join_playlist = make_join_playlist(vods_m3u8, vods, targets)
    join_playlist_path = cache.get_path("playlist_downloaded.m3u8")
    join_playlist.dump(join_playlist_path)  # type: ignore
    click.echo()

    if args.no_join:
        print_log("Skipping joining files...")
        click.echo(f"VODs downloaded to:\n{blue(cache_dir)}")
        return

    if args.concat:
        print_log("Concating files...")
        _concat_vods(targets, target)
    else:
        print_log("Joining files...")
        _join_vods(
            join_playlist_path,
            metadata_path,
            target,
            overwrite,
            crop_start,
            crop_duration,
        )

    click.echo()

    if args.keep:
        click.echo(f"Cached files not deleted: {yellow(cache_dir)}")
    else:
        print_log("Deleting cached files...")
        cache.delete()

    click.echo(f"Downloaded: {green(target)}")


def _get_cache_dir(video: Video, playlist: Playlist, options: DownloadOptions) -> Path:
    subs = video_placeholders(video, options.format)
    subs["quality"] = playlist.group_id

    try:
        return Path(options.cache_dir.format(**subs))
    except KeyError as e:
        supported = ", ".join(subs.keys())
        raise ConsoleError(f"Invalid key {e} used in --cache-dir. Supported keys are: {supported}")


def _print_found_video(video: Video):
    print_log(
        "Found video:",
        green(video["title"]),
        "by",
        yellow(video["owner"]["displayName"]),
        "playing",
        blue(video["game"]["name"] if video["game"] else "Unknown"),
        f"({utils.format_time(video['lengthSeconds'])})",
    )


def http_get(url: str) -> str:
    response = httpx.get(url)
    response.raise_for_status()
    return response.text


def _determine_time_range(chapters: List[Chapter], args: DownloadOptions):
    if args.start or args.end:
        return args.start, args.end

    if args.chapter is not None:
        if not chapters:
            raise ConsoleError("This video has no chapters")

        if args.chapter == 0:
            chapter = _choose_chapter_interactive(chapters)
        else:
            try:
                chapter = chapters[args.chapter - 1]
            except IndexError:
                raise ConsoleError(
                    f"Chapter {args.chapter} does not exist. This video has {len(chapters)} chapters."
                )

        click.echo(f'Chapter selected: {blue(chapter["description"])}\n')
        start = chapter["positionMilliseconds"] // 1000
        duration = chapter["durationMilliseconds"] // 1000
        return start, start + duration

    return None, None


def _choose_chapter_interactive(chapters: List[Chapter]):
    click.echo("\nChapters:")
    for index, chapter in enumerate(chapters):
        duration = utils.format_time(chapter["durationMilliseconds"] // 1000)
        click.echo(f'{index + 1}) {bold(chapter["description"])} ({duration})')
    index = utils.read_int("Select a chapter", 1, len(chapters))
    chapter = chapters[index - 1]
    return chapter


# See: https://ffmpeg.org/ffmpeg-formats.html#Metadata-2
def _write_metadata(
    video: Video,
    chapters: List[Chapter],
    path: Path,
    start_offset: Optional[int],
    end_offset: Optional[int],
):
    start_offset_ms = 1000 * (start_offset or 0)
    end_offset_ms = 1000 * (end_offset or video["lengthSeconds"])

    date = _escape_metadata(video["publishedAt"][:10])
    title = _escape_metadata(video["title"])
    artist = _escape_metadata(video["owner"]["displayName"])
    description = _escape_metadata(video["description"])
    show = _escape_metadata(video["game"]["name"]) if video["game"] else None

    with open(path, "w", encoding="utf-8") as f:
        # Header
        f.write(";FFMETADATA1\n")

        # Global metadata
        f.write(f"date={date}\n")
        f.write(f"title={title}\n")
        f.write(f"artist={artist}\n")
        if description:
            f.write(f"description={description}\n")
        if show:
            f.write(f"show={show}\n")
        f.write("encoded_by=twitch-dl\n")

        # Chapter metadata
        for chapter in chapters:
            title = chapter["description"]
            chapter_start_ms = chapter["positionMilliseconds"]
            chapter_end_ms = chapter["positionMilliseconds"] + chapter["durationMilliseconds"]

            if chapter_start_ms >= start_offset_ms and chapter_start_ms <= end_offset_ms:
                if chapter_end_ms > end_offset_ms:
                    chapter_end_ms = end_offset_ms

                start_ms = chapter_start_ms - start_offset_ms
                end_ms = chapter_end_ms - start_offset_ms

                f.write("\n[CHAPTER]\n")
                f.write("TIMEBASE=1/1000\n")
                f.write(f"START={start_ms}\n")
                f.write(f"END={end_ms}\n")
                f.write(f"title={title}\n")


def _escape_metadata(text: Optional[str]):
    #  Metadata keys or values containing special characters
    # (‘=’, ‘;’, ‘#’, ‘\’ and a newline) must be escaped with a backslash ‘\’.
    text = text.strip() if text else ""
    return re.sub(r"([=;#\\\n])", r"\\\1", text.strip())


class Overwrite(Enum):
    OVERWRITE = auto()
    OVERWRITE_ALL = auto()
    SKIP = auto()
    SKIP_ALL = auto()
    ABORT = auto()


def _prompt_overwrite() -> Overwrite:
    prompt = (
        "File exists. Do you want to: "
        + f"{underlined('O')}verwrite, "
        + f"Overwrite {underlined('A')}ll, "
        + f"{underlined('S')}kip, "
        + f"S{underlined('k')}ip all, "
        + f"A{underlined('b')}ort"
    )

    while True:
        response = click.prompt(prompt, default="O").lower().strip()

        if response == "o":
            return Overwrite.OVERWRITE
        elif response == "a":
            return Overwrite.OVERWRITE_ALL
        elif response == "s":
            return Overwrite.SKIP
        elif response == "k":
            return Overwrite.SKIP_ALL
        elif response == "b":
            return Overwrite.ABORT

        print_error(f"Invalid response: {response}")



import logging
import platform
import re
import shutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent
from typing import List, Optional, Tuple

import click

from twitchdl import __version__
from twitchdl.cache import get_cache_dir, get_cache_subdirs
from twitchdl.entities import DownloadOptions
from twitchdl.exceptions import ConsoleError
from twitchdl.naming import DEFAULT_OUTPUT_TEMPLATE
from twitchdl.output import print_log, print_table
from twitchdl.twitch import ClipsPeriod, VideosSort, VideosType
from twitchdl.utils import format_size, get_size

# Tweak the Click context
# https://click.palletsprojects.com/en/8.1.x/api/#context
CONTEXT = dict(
    # Enable using environment variables to set options
    auto_envvar_prefix="TWITCH_DL",
    # Add shorthand -h for invoking help
    help_option_names=["-h", "--help"],
    # Always show default values for options
    show_default=True,
    # Make help a bit wider
    max_content_width=100,
)

json_option = click.option(
    "--json",
    is_flag=True,
    default=False,
    help="Print data as JSON rather than human readable text",
)


def validate_positive(_ctx: click.Context, _param: click.Parameter, value: Optional[int]):
    if value is not None and value <= 0:
        raise click.BadParameter("must be greater than 0")
    return value


def validate_non_negative(_ctx: click.Context, _param: click.Parameter, value: Optional[int]):
    if value is not None and value < 0:
        raise click.BadParameter("must be greater or equal than 0")
    return value


def validate_time(_ctx: click.Context, _param: click.Parameter, value: str) -> Optional[int]:
    """Parse a time string (hh:mm or hh:mm:ss) to number of seconds."""
    if not value:
        return None

    parts = [int(p) for p in value.split(":")]

    if not 2 <= len(parts) <= 3:
        raise click.BadParameter("invalid time")

    hours = parts[0]
    minutes = parts[1]
    seconds = parts[2] if len(parts) > 2 else 0

    if hours < 0 or not (0 <= minutes <= 59) or not (0 <= seconds <= 59):
        raise click.BadParameter("invalid time")

    return hours * 3600 + minutes * 60 + seconds


def validate_rate(_ctx: click.Context, _param: click.Parameter, value: str) -> Optional[int]:
    if not value:
        return None

    match = re.search(r"^([0-9]+)(k|m|)$", value, flags=re.IGNORECASE)

    if not match:
        raise click.BadParameter("must be an integer, followed by an optional 'k' or 'm'")

    amount = int(match.group(1))
    unit = match.group(2)

    if unit == "k":
        return amount * 1024

    if unit == "m":
        return amount * 1024 * 1024

    return amount


@click.group(context_settings=CONTEXT)
@click.option("--debug/--no-debug", default=False, help="Enable debug logging to stderr")
@click.option("--verbose/--no-verbose", default=False, help="More verbose debug logging")
@click.option("--color/--no-color", default=sys.stdout.isatty(), help="Use ANSI color in output")
@click.version_option(package_name="twitch-dl")
@click.pass_context
def cli(ctx: click.Context, color: bool, debug: bool, verbose: bool):
    """twitch-dl - twitch.tv downloader

    https://twitch-dl.bezdomni.net/
    """
    ctx.color = color

    if debug:
        logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
        logging.getLogger("httpx").setLevel(logging.WARN)
        logging.getLogger("httpcore").setLevel(logging.WARN)
        logging.getLogger("PIL").setLevel(logging.WARN)


@cli.command()
@click.argument("channel_name")
@click.option(
    "-a",
    "--all",
    help="Fetch all clips, overrides --limit",
    is_flag=True,
)
@click.option(
    "-c",
    "--compact",
    help="Show clips in compact mode, one line per video",
    is_flag=True,
)
@click.option(
    "-d",
    "--download",
    help="Download clips in given period (in source quality)",
    is_flag=True,
)
@click.option(
    "-l",
    "--limit",
    help="Number of clips to fetch. Defaults to 40 in compact mode, 10 otherwise.",
    type=int,
    callback=validate_positive,
)
@click.option(
    "-p",
    "--pager",
    help="Number of clips to show per page. Disabled by default.",
    type=int,
    callback=validate_positive,
    is_flag=False,
    flag_value=10,
)
@click.option(
    "-P",
    "--period",
    help="Period from which to return clips",
    default="all_time",
    type=click.Choice(["last_day", "last_week", "last_month", "all_time"]),
)
@click.option(
    "-t",
    "--target-dir",
    help="Target directory when downloading clips",
    type=click.Path(
        file_okay=False,
        readable=False,
        writable=True,
        path_type=Path,
    ),
    default=Path(),
)
@json_option
def clips(
    channel_name: str,
    all: bool,
    compact: bool,
    download: bool,
    json: bool,
    limit: Optional[int],
    pager: Optional[int],
    period: ClipsPeriod,
    target_dir: Path,
):
    """List or download clips for given CHANNEL_NAME."""
    from twitchdl.commands.clips import clips

    if not target_dir.exists():
        target_dir.mkdir(parents=True, exist_ok=True)

    clips(
        channel_name,
        all=all,
        compact=compact,
        download=download,
        json=json,
        limit=limit,
        pager=pager,
        period=period,
        target_dir=target_dir,
    )


@cli.command()
@click.argument("ids", nargs=-1)
@click.option(
    "-a",
    "--auth-token",
    help="""Authentication token, passed to Twitch to access subscriber only
         VODs. Can be copied from the `auth_token` cookie in any browser logged
         in on Twitch.""",
)
@click.option(
    "-c",
    "--chapter",
    help="""Download a single chapter of the video. Specify the chapter number
         or use the flag without a number to display a chapter select prompt.
         """,
    type=int,
    is_flag=False,
    flag_value=0,
)
@click.option(
    "--concat",
    is_flag=True,
    help="""Do not use ffmpeg to join files, concat them instead. This will
         produce a .ts file by default.""",
)
@click.option(
    "-d",
    "--dry-run",
    help="Simulate the download provcess without actually downloading any files.",
    is_flag=True,
)
@click.option(
    "-e",
    "--end",
    help="Download video up to this time (hh:mm or hh:mm:ss)",
    callback=validate_time,
)
@click.option(
    "-f",
    "--format",
    help="""Video format to convert into, passed to ffmpeg as the target file
         extension. Defaults to `mkv`. If `--concat` is passed, defaults to
         `ts`.""",
)
@click.option(
    "-k",
    "--keep",
    help="Don't delete downloaded VODs and playlists after merging.",
    is_flag=True,
)
@click.option(
    "--no-join",
    help="Don't run ffmpeg to join the downloaded vods, implies --keep.",
    is_flag=True,
)
@click.option(
    "--overwrite",
    help="Overwrite target file if it already exists",
    is_flag=True,
)
@click.option(
    "--skip-existing",
    help="Skip target file if it already exists",
    is_flag=True,
)
@click.option(
    "-o",
    "--output",
    help="Output file name template. See docs for details.",
    default=DEFAULT_OUTPUT_TEMPLATE,
)
@click.option(
    "-q",
    "--quality",
    help="Video quality, e.g. `720p`. Set to `source` to get best quality.",
)
@click.option(
    "-r",
    "--rate-limit",
    help="""Limit the maximum download speed in bytes per second. Use 'k' and
         'm' suffixes for kbps and mbps.""",
    callback=validate_rate,
)
@click.option(
    "-s",
    "--start",
    help="Download video from this time (hh:mm or hh:mm:ss)",
    callback=validate_time,
)
@click.option(
    "-w",
    "--max-workers",
    help="Number of workers for downloading vods concurrently",
    type=int,
    default=10,
)
@click.option(
    "--cache-dir",
    help="Folder where VODs are downloaded before joining. Uses placeholders similar to --output.",
    default=f"{get_cache_dir()}/videos/{{id}}/{{quality}}",
)
def download(
    ids: Tuple[str, ...],
    auth_token: Optional[str],
    chapter: Optional[int],
    concat: bool,
    dry_run: bool,
    end: Optional[int],
    format: str,
    keep: bool,
    no_join: bool,
    overwrite: bool,
    skip_existing: bool,
    output: str,
    quality: Optional[str],
    rate_limit: Optional[int],
    start: Optional[int],
    max_workers: int,
    cache_dir: str,
):
    """Download videos or clips.

    Pass one or more video ID, clip slug or Twitch URL to download.
    """
    from twitchdl.commands.download import download

    if not format:
        format = "ts" if concat else "mkv"

    if start is not None and end is not None and end <= start:
        raise ConsoleError("End time must be greater than start time")

    options = DownloadOptions(
        auth_token=auth_token,
        chapter=chapter,
        concat=concat,
        dry_run=dry_run,
        end=end,
        format=format,
        keep=keep,
        no_join=no_join,
        overwrite=overwrite,
        skip_existing=skip_existing,
        output=output,
        quality=quality,
        rate_limit=rate_limit,
        start=start,
        max_workers=max_workers,
        cache_dir=cache_dir,
    )

    download(list(ids), options)


@cli.command()
def env():
    """Print environment information for inclusion in bug reports."""
    click.echo(f"twitch-dl {__version__}")
    click.echo(f"Python {sys.version}")
    click.echo(f"Platform: {platform.platform()}")

    click.echo("--")
    ffmpeg = shutil.which("ffmpeg")
    if ffmpeg:
        click.echo(f"ffmpeg path: {ffmpeg}")
        try:
            subprocess.run(["ffmpeg", "-version"])
        except Exception as ex:
            click.secho("Failed getting ffmpeg version:", fg="red")
            click.secho(f"{ex}", fg="red")
    else:
        click.secho("ffmpeg not found", err=True, fg="red")


@cli.command()
@click.argument("id")
@json_option
def info(id: str, json: bool):
    """Print information for a given Twitch URL, video ID or clip slug."""
    from twitchdl.commands.info import info

    info(id, json=json)


@cli.command()
@click.argument("channel_name")
@click.option(
    "-a",
    "--all",
    help="Fetch all clips, overrides --limit",
    is_flag=True,
)
@click.option(
    "-c",
    "--compact",
    help="Show videos in compact mode, one line per video",
    is_flag=True,
)
@click.option(
    "-l",
    "--limit",
    help="Number of videos to fetch. Defaults to 40 in compact mode, 10 otherwise.",
    type=int,
    callback=validate_positive,
)
@click.option(
    "-p",
    "--pager",
    help="Number of videos to show per page. Disabled by default.",
    type=int,
    callback=validate_positive,
    is_flag=False,
    flag_value=10,
)
@click.option(
    "-g",
    "--game",
    "games_tuple",
    help="Show videos of given game (can be given multiple times)",
    multiple=True,
)
@click.option(
    "-s",
    "--sort",
    help="Sorting order of videos",
    default="time",
    type=click.Choice(["views", "time"]),
)
@click.option(
    "-t",
    "--type",
    help="Broadcast type",
    default="archive",
    type=click.Choice(["archive", "highlight", "upload"]),
)
@json_option
def videos(
    channel_name: str,
    all: bool,
    compact: bool,
    games_tuple: Tuple[str, ...],
    json: bool,
    limit: Optional[int],
    pager: Optional[int],
    sort: VideosSort,
    type: VideosType,
):
    """List or download clips for given CHANNEL_NAME."""
    from twitchdl.commands.videos import videos

    # Click provides a tuple, make it a list instead
    games = list(games_tuple)

    videos(
        channel_name,
        all=all,
        compact=compact,
        games=games,
        json=json,
        limit=limit,
        pager=pager,
        sort=sort,
        type=type,
    )


@cli.command()
@click.argument("id")
@click.option(
    "-w",
    "--width",
    help="Chat width in pixels",
    type=int,
    default=400,
    callback=validate_positive,
)
@click.option(
    "-h",
    "--height",
    help="Chat height in pixels",
    type=int,
    default=1080,
    callback=validate_positive,
)
@click.option(
    "--font-size",
    help="Font size",
    type=int,
    default=20,
    callback=validate_positive,
)
@click.option(
    "--dark",
    help="Dark mode",
    is_flag=True,
)
@click.option(
    "--pad-x",
    help="Horizontal padding",
    type=int,
    callback=validate_non_negative,
    default=5,
)
@click.option(
    "--pad-y",
    help="Vertical padding",
    type=int,
    callback=validate_non_negative,
    default=5,
)
@click.option(
    "-o",
    "--output",
    help="Output file name template. See docs for details.",
    default="chat_{id}.{format}",
)
@click.option(
    "-f",
    "--format",
    help="Video format to convert into, passed to ffmpeg as the target file extension.",
    default="mp4",
)
@click.option(
    "--overwrite",
    help="Overwrite the target file if it already exists without prompting.",
    is_flag=True,
)
@json_option
def chat(
    id: str,
    width: int,
    height: int,
    font_size: int,
    dark: bool,
    pad_x: int,
    pad_y: int,
    output: str,
    format: str,
    overwrite: bool,
    json: bool,
):
    """
    Render chat for a given video.

    This command is experimental and may change in the future!
    """
    print_log("Chat command is still experimental, try it out and report any bugs.")

    try:
        from twitchdl.chat import render_chat

        render_chat(
            id,
            width,
            height,
            font_size,
            dark,
            (pad_x, pad_y),
            output,
            format,
            overwrite,
            json,
        )
    except ModuleNotFoundError as ex:
        raise ConsoleError(
            dedent(f"""
                {ex}

                This command requires twitch-dl to be installed with optional "chat" dependencies:
                pipx install "twitch-dl[chat]"

                See documentation for more info:
                https://twitch-dl.bezdomni.net/commands/chat.html
            """)
        )


@cli.command
@click.option(
    "-c",
    "--clear",
    "clear_subdir",
    help="Clear cached files",
    type=click.Choice(["all", "fonts", "chats", "videos", "emotes", "badges"]),
)
def cache(clear_subdir: str):
    """View and manage cached files"""
    if clear_subdir:
        clear_path = get_cache_dir() if clear_subdir == "all" else get_cache_dir(clear_subdir)
        size = get_size(clear_path)
        shutil.rmtree(clear_path)
        click.echo(f"Cleared {clear_subdir} cache ({format_size(size)})")
        return

    cache_dir = get_cache_dir()
    click.echo(f"Cache dir: {cache_dir}")

    rows: List[List[str]] = []
    total_size = 0
    for directory in get_cache_subdirs():
        size = get_size(directory)
        rows.append([str(directory), format_size(size)])
        total_size += size

    if not rows:
        click.echo("No files cached")
        return

    click.echo()
    print_table(
        rows,
        headers=["Directory", "Size"],
        footers=["Total", format_size(total_size)],
        alignments={1: "right"},
    )
